@@ -1,8 +1,5 @@
 # -*- coding: utf-8 -*-
 
-# FIXME python2
-from __future__ import unicode_literals
-
 import logging
 import os
 import shutil
@@ -12,15 +9,15 @@ from datetime import datetime, timedelta
 
 from dateutil import tz
 from lxml.etree import Element
+from pathlib import Path
 
 from io import BytesIO
 
 from pykeepass import PyKeePass, icons
 from pykeepass.entry import Entry
-from pykeepass.exceptions import BinaryError
 from pykeepass.group import Group
 from pykeepass.kdbx_parsing import KDBX
-from pykeepass.exceptions import BinaryError, CredentialsError
+from pykeepass.exceptions import BinaryError, CredentialsError, HeaderChecksumError
 
 """
 Missing Tests:
@@ -197,6 +194,8 @@ class EntryFindTests3(KDBX3Tests):
         hist = entry.history
         self.assertIsInstance(hist, list)
         self.assertEqual(len(hist), 4)
+        self.assertEqual(len(set(hist)), 4)
+        self.assertNotEqual(hist[0], hist[1])
 
     def test_history_path(self):
         for title in ["root_entry", "subentry"]:
@@ -228,7 +227,7 @@ class EntryFindTests3(KDBX3Tests):
             unique_str + 'pass',
             url=unique_str + 'url',
             notes=unique_str + 'notes',
-            tags=unique_str + 'tags',
+            tags=['tag1', 'tag2', 'tag3;tag4', 'tag5,tag6'],
             expiry_time=expiry_time,
             icon=icons.KEY
         )
@@ -241,7 +240,7 @@ class EntryFindTests3(KDBX3Tests):
         self.assertEqual(results.password, unique_str + 'pass')
         self.assertEqual(results.url, unique_str + 'url')
         self.assertEqual(results.notes, unique_str + 'notes')
-        self.assertEqual(results.tags, [unique_str + 'tags'])
+        self.assertEqual(len(results.tags), 6)
         self.assertTrue(results.uuid != None)
         self.assertTrue(results.autotype_sequence is None)
         # convert naive datetime to utc
@@ -469,6 +468,7 @@ class EntryTests3(KDBX3Tests):
 
     def test_references(self):
         original_entry = self.kp.find_entries(title='foobar_entry', first=True)
+        original_entry_duplicate = self.kp.find_entries(title='foobar_entry', first=True)
         clone1 = self.kp.find_entries(title='foobar_entry - Clone', first=True)
         clone2 = self.kp.find_entries(title='foobar_entry - Clone of clone', first=True)
         prefixed = self.kp.find_entries(title='foobar_entry - Clone with prefix and suffix', first=True)
@@ -478,6 +478,10 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(original_entry.ref('username'), clone1.username)
         self.assertEqual(prefixed.deref('username'), 'domain\\{}2'.format(original_entry.username))
         self.assertEqual(prefixed.deref('password'), 'A{}BC'.format(original_entry.password))
+        self.assertEqual(original_entry, original_entry_duplicate)
+        self.assertEqual(hash(original_entry), hash(original_entry_duplicate))
+        self.assertNotEqual(original_entry, clone1)
+        self.assertNotEqual(clone1, clone2)
 
     def test_set_and_get_fields(self):
         time = datetime.now().replace(microsecond=0)
@@ -597,6 +601,16 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(len(entry.attachments), num_attach + 1)
         self.assertEqual(entry.attachments[0].filename, 'foobar2.txt')
 
+    def test_is_custom_property_protected(self):
+        e = self.kp.add_entry(self.kp.root_group, 'test-protect', 'some-user', 'pass')
+        e.set_custom_property('protected', 'something', protect=True)
+        e.set_custom_property('explicit-unprotected', 'other', protect=False)
+        e.set_custom_property('not-protected', 'secret')
+        self.assertTrue(e.is_custom_property_protected('protected'))
+        self.assertFalse(e.is_custom_property_protected('explicit-unprotected'))
+        self.assertFalse(e.is_custom_property_protected('not-protected'))
+        self.assertFalse(e.is_custom_property_protected('non-existent'))
+
 
 class EntryHistoryTests3(KDBX3Tests):
 
@@ -769,7 +783,7 @@ class AttachmentTests3(KDBX3Tests):
         self.kp_tmp.save()
         self.kp_tmp.reload()
         self.assertEqual(self.kp_tmp.binaries[binary_id], b'Ronald McDonald Trump')
-        self.assertEqual(len(self.kp_tmp.attachments), 1)
+        self.assertEqual(len(self.kp_tmp.attachments), 2)
 
         num_attach = len(self.kp_tmp.binaries)
         self.kp_tmp.delete_binary(binary_id)
@@ -777,6 +791,11 @@ class AttachmentTests3(KDBX3Tests):
         self.kp_tmp.reload()
         self.assertEqual(len(self.kp_tmp.binaries), num_attach - 1)
 
+        # test empty attachment - issue 314
+        a = self.kp.find_attachments(filename='foo.txt', first=True)
+        self.assertEqual(a._element.text, None)
+        self.assertEqual(a.data, b'')
+
     def test_attachment_reference_decrement(self):
         e = self.kp.entries[0]
 
@@ -791,6 +810,7 @@ class AttachmentTests3(KDBX3Tests):
         self.assertEqual(attachment2.id, binary_id2 - 1)
 
     def test_fields(self):
+        # test creation
         e = self.kp.entries[0]
         binary_id = self.kp.add_binary(b'foobar')
         a = e.add_attachment(filename='test.txt', id=binary_id)
@@ -937,6 +957,11 @@ class BugRegressionTests3(KDBX3Tests):
 
         self.assertEqual(results, results2)
 
+    def test_issue344(self):
+        # accessing expiry_time throws exception when None
+
+        e = self.kp.find_entries(title='none_date', first=True)
+        self.assertEqual(e.expiry_time, None)
 
 class EntryFindTests4(KDBX4Tests, EntryFindTests3):
     pass
@@ -980,7 +1005,7 @@ class KDBXTests(unittest.TestCase):
 
         filenames_in = [
             os.path.join(base_dir, 'test3.kdbx'),                 # KDBX v3
-            os.path.join(base_dir, 'test4.kdbx'),                 # KDBX v4
+            Path(base_dir).joinpath('test4.kdbx'),                # KDBX v4 (and test pathlib)
             os.path.join(base_dir, 'test4_aes.kdbx'),             # KDBX v4 AES
             os.path.join(base_dir, 'test4_aeskdf.kdbx'),          # KDBX v3 AESKDF
             os.path.join(base_dir, 'test4_chacha20.kdbx'),        # KDBX v4 ChaCha
@@ -988,7 +1013,7 @@ class KDBXTests(unittest.TestCase):
             os.path.join(base_dir, 'test4_hex.kdbx'),             # legacy 64 byte hexadecimal keyfile
             os.path.join(base_dir, 'test3_transformed.kdbx'),     # KDBX v3 transformed_key open
             os.path.join(base_dir, 'test4_transformed.kdbx'),     # KDBX v4 transformed_key open
-            stream,
+            stream,                                               # test stream opening
             os.path.join(base_dir, 'test4_aes_uncompressed.kdbx'),# KDBX v4 AES uncompressed
             os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx'),# KDBX v4 Twofish uncompressed
             os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx'),# KDBX v4 ChaCha uncompressed
@@ -996,7 +1021,7 @@ class KDBXTests(unittest.TestCase):
         ]
         filenames_out = [
             os.path.join(base_dir, 'test3.kdbx.out'),
-            os.path.join(base_dir, 'test4.kdbx.out'),
+            Path(base_dir).joinpath('test4.kdbx.out'),
             os.path.join(base_dir, 'test4_aes.kdbx.out'),
             os.path.join(base_dir, 'test4_aeskdf.kdbx.out'),
             os.path.join(base_dir, 'test4_chacha20.kdbx.out'),
@@ -1044,7 +1069,7 @@ class KDBXTests(unittest.TestCase):
         ]
         keyfiles = [
             'test3.key',
-            'test4.key',
+            Path('test4.key'),
             'test4.key',
             'test4.key',
             'test4.key',
@@ -1143,28 +1168,38 @@ class KDBXTests(unittest.TestCase):
                 os.remove(os.path.join(base_dir, filename))
 
 
-    def test_credentials_error(self):
+    def test_open_error(self):
 
         databases = [
             'test3.kdbx',
             'test3.kdbx',
             'test4.kdbx',
-            'test4.kdbx'
+            'test4.kdbx',
+            'test3.key',
         ]
         passwords = [
             'invalid',
             'password',
             'invalid',
             'password',
+            'password',
         ]
         keyfiles = [
             'test3.key',
             'test4.key',
             'test4.key',
             'test3.key',
+            'test3.key',
         ]
-        for database, password, keyfile in zip(databases, passwords, keyfiles):
-            with self.assertRaises(CredentialsError):
+        errors = [
+            CredentialsError,
+            CredentialsError,
+            CredentialsError,
+            CredentialsError,
+            HeaderChecksumError,
+        ]
+        for database, password, keyfile, error in zip(databases, passwords, keyfiles, errors):
+            with self.assertRaises(error):
                 PyKeePass(
                     os.path.join(base_dir, database),
                     password,
